/*
 * Copyright 2020-2021 David Sugar, Tycho Softworks.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "sessions.hpp"
#include "logger.hpp"
#include "media.hpp"

#include <exosip.hpp>

#include <functional>

namespace server {
using segment_t = struct Segment final : public server::Session { // NOLINT
    Segment(script& attach, const exosip::event_t& event) :
    Session(attach), cid(event.cid()), did(event.did()), tid(event.tid()), context(event.context()) {}

    inline auto make_const(const symbols::init_t& list) {
        return syms.make_const(list);
    }

    inline auto ipc() const {
        return shm;
    }

    auto sdp() {
        if(!audio || !audio->local_port)
            return std::string();

        auto uid = syms.get_value("CALLED_TO");
        auto ip = ipv6 ? "IP6" : "IP4";
        if(!audio->dtmf) {
            return print_format(
                "v=0\r\n"
                "o={} 0 0 IN {} {}\r\n"
                "s={}\r\n"
                "c=IN {} {}\r\n"
                "t=0 0\r\n"
                "m=audio {} RTP/AVP {}\r\n"
                "a=rtpmap:{} pcmu/8000\r\n",
                uid, ip, audio->media,
                sid(),
                ip, audio->media,
                audio->local_port, audio->payload,
                audio->payload
            );
        }
        return print_format(
            "v=0\r\n"
            "o={} 0 0 IN {} {}\r\n"
            "s={}\r\n"
            "c=IN {} {}\r\n"
            "t=0 0\r\n"
            "m=audio {} RTP/AVP {} {}\r\n"
            "a=rtpmap:{} pcmu/8000\r\n"
            "a=rtpmap:{} telephone-event/8000\r\n",
            uid, ip, audio->media,
            sid(),
            ip, audio->media,
            audio->local_port, audio->payload, audio->dtmf,
            audio->payload,
            audio->dtmf
        );
    }

    auto answer(int status, std::function<void(exosip::msg_t)> fill = [](auto msg){}) {
        if(tid < 1)
            return false;

        auto msg = exosip::msg_t(nullptr);
        exosip::guard_t lock(context);
        eXosip_call_build_answer(context, tid, status, &msg);
        if(!msg)
            return false;

        fill(msg);
        eXosip_call_send_answer(context, tid, status, msg);
        if(status >= 200)
            tid = -1;

        return true;
    }

    void terminate() {
        if(tid > 0)
            answer(SIP_REQUEST_TERMINATED);
        else if(cid > 0 && did > 0) {
            exosip::guard_t lock(context);
            eXosip_call_terminate(context, cid, did);
        }
        cid = did = tid = -1;
        connected = false;
    }

    int cid, did, tid;
    eXosip_t *context;
};
}
