/*
 * Copyright (C) 2020-2021 David Sugar <tychosoft@gmail.com>.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "clock.hpp"
#include "drivers.hpp"

#include <thread>
#include <string>
#include <cstring>

#include <pthread.h>
#include <sched.h>

using namespace server;

void util::timespec_delay(timespec *current, milliseconds_t offset)
{
    current->tv_sec += offset / 1000U;
    current->tv_nsec += static_cast<long>(offset % 1000L) * 1000000L;
    while(current->tv_nsec >= 1000000000U) {
        ++current->tv_sec;
        current->tv_nsec -= 1000000000U;
    }
}

void util::sleep(milliseconds_t msec)
{
    auto timeout = std::chrono::milliseconds(msec);
    std::this_thread::sleep_for(timeout);
}

void util::sleep(const steady_time& until)
{
    auto timeout = util::steady_remains(until);
    if(timeout.count() <= 0)
        return;

    std::this_thread::sleep_for(timeout);
}

auto util::sleep_until(const timer& until) -> bool
{
    auto duration = until - timer();
    if(!duration)
        return false;
#if defined(TIMER_ABSTIME) && !defined(__OpenBSD__)
    clock_nanosleep(CLOCK_MONOTONIC, TIMER_ABSTIME, &until, nullptr);
#else
    nanosleep(&duration, nullptr);
#endif
    return true;
}

timer::timer(milliseconds_t msec) : timespec()
{
    clock_gettime(CLOCK_MONOTONIC, this);
    util::timespec_delay(this, msec);
}

auto timer::operator=(milliseconds_t msec) -> timer&
{
    clock_gettime(CLOCK_MONOTONIC, this);
    util::timespec_delay(this, msec);
    return *this;
}

auto timer::operator+=(milliseconds_t msec) -> timer&
{
    util::timespec_delay(this, msec);
    return *this;
}

auto timer::operator-=(const timer& base) -> timer&
{
    if(base >= *this) {
        tv_sec = tv_nsec = 0;
    }
    else {
        tv_sec -= base.tv_sec;
        if(tv_nsec < base.tv_nsec) {
            --tv_sec;
            tv_nsec += 1000000000U;
        }
        tv_nsec -= base.tv_nsec;
    }
    return *this;
}

auto timer::until() const -> milliseconds_t
{
    auto change = *this - timer();
    return static_cast<milliseconds_t>(change.tv_sec * 1000U + change.tv_nsec / 1000000U);
}

auto server::operator<(const timer& lh, const timer& rh) -> bool
{
    if(lh.tv_sec < rh.tv_sec)
        return true;
    if(lh.tv_sec > rh.tv_sec)
        return false;
    return lh.tv_nsec < rh.tv_nsec;
}

auto server::operator>(const timer& lh, const timer& rh) -> bool
{
    if(lh.tv_sec > rh.tv_sec)
        return true;
    if(lh.tv_sec < rh.tv_sec)
        return false;
    return lh.tv_nsec > rh.tv_nsec;
}

auto server::operator<=(const timer& lh, const timer& rh) -> bool
{
    if(lh.tv_sec < rh.tv_sec)
        return true;
    if(lh.tv_sec > rh.tv_sec)
        return false;
    return lh.tv_nsec <= rh.tv_nsec;
}

auto server::operator>=(const timer& lh, const timer& rh) -> bool
{
    if(lh.tv_sec > rh.tv_sec)
        return true;
    if(lh.tv_sec < rh.tv_sec)
        return false;
    return lh.tv_nsec >= rh.tv_nsec;
}

auto util::to_duration(const char *cp) -> milliseconds_t
{
    char *ep = nullptr;
    if(!cp)
        return 0;

    auto duration = static_cast<milliseconds_t>(strtol(cp, &ep, 10));

    if(!ep || !*ep || !strcmp(ep, "s"))
        return duration * 1000L;

    if(!strcmp(ep, "r"))    // ring timed interval...
        return  static_cast<unsigned long>(duration) * Driver::ring_timer() * 1000L;

    if(!strcmp(ep, "m"))
        return duration * 60000L;

    if(!strcmp(ep, "ms"))
        return duration;

    return 0;
}

