/*
 * Copyright (C) 2020-2021 David Sugar <tychosoft@gmail.com>.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "drivers.hpp"
#include "runtime.hpp"
#include "logger.hpp"
#include "utils.hpp"
#include "proc.hpp"

#include <strings.hpp>
#include <print.hpp>
#include <keyfile.hpp>

#include <fstream>
#include <list>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <unistd.h>
#include <fcntl.h>

using namespace server;

namespace  {
Session::event_t shutdown_event{event_type::shutdown};
const char *dtmf_group = "dtmf";

std::list<ipc_session*> ipc_free;
ipc_session *ipc_mapped = nullptr;
keyfile::keys timers;
} // end namespace

Plugin *Plugin::plugins_ = nullptr;
Driver *Driver::driver_ = nullptr;
ipc_system *Driver::sys_mapped = nullptr;
unsigned Driver::ring_timer_ = 4;
unsigned Driver::answer_timer_ = 30;
std::unordered_map<std::string, script> Driver::scripts;
std::unordered_map<std::string_view,session_t> Driver::sessions;
std::mutex Driver::session_lock;
std::mutex Driver::script_lock;

std::unordered_map<std::string_view, const char *> Driver::groups = {
    {"digits", dtmf_group},
    {"1", dtmf_group},
    {"2", dtmf_group},
    {"3", dtmf_group},
    {"4", dtmf_group},
    {"5", dtmf_group},
    {"6", dtmf_group},
    {"7", dtmf_group},
    {"8", dtmf_group},
    {"9", dtmf_group},
    {"*", dtmf_group},
    {"#", dtmf_group},
};

std::unordered_map<std::string_view, Driver::keyword_t> Driver::keywords = {
    {"debug", {
        static_cast<method_t>(&methods::scr_debug),
        static_cast<check_t>(&checks::chk_debug)}},
    {"pause", {
        static_cast<method_t>(&methods::scr_pause),
        static_cast<check_t>(&checks::chk_keyword)}},
    {"exit", {
        static_cast<method_t>(&methods::scr_exit),
        static_cast<check_t>(&checks::chk_keyword)}},
    {"event", {
        static_cast<method_t>(&methods::scr_event),
        static_cast<check_t>(&checks::chk_event)}},
    {"do", {
        static_cast<method_t>(&methods::scr_do),
        static_cast<check_t>(&checks::chk_do)}},
    {"repeat", {
        static_cast<method_t>(&methods::scr_exit),
        static_cast<check_t>(&checks::chk_conditional)}},
    {"break", {
        static_cast<method_t>(&methods::scr_break),
        static_cast<check_t>(&checks::chk_conditional)}},
    {"end", {
        static_cast<method_t>(&methods::scr_end),
        static_cast<check_t>(&checks::chk_conditional)}},
    {"switch", {
        static_cast<method_t>(&methods::scr_switch),
        static_cast<check_t>(&checks::chk_conditional)}},
    {"case", {
        static_cast<method_t>(&methods::scr_case),
        static_cast<check_t>(&checks::chk_case)}},
    {"otherwise", {
        static_cast<method_t>(&methods::scr_otherwise),
        static_cast<check_t>(&checks::chk_otherwise)}},
    {"var", {
        static_cast<method_t>(&methods::scr_var),
        static_cast<check_t>(&checks::chk_var)}},
    {"set", {
        static_cast<method_t>(&methods::scr_set),
        static_cast<check_t>(&checks::chk_set)}},
    {"const", {
        static_cast<method_t>(&methods::scr_const),
        static_cast<check_t>(&checks::chk_const)}},
    {"sleep", {
        static_cast<method_t>(&methods::scr_sleep),
        static_cast<check_t>(&checks::chk_sleep)}},
    {"until", {
        static_cast<method_t>(&methods::scr_until),
        static_cast<check_t>(&checks::chk_sleep)}},
    {"clear", {
        static_cast<method_t>(&methods::scr_clear),
        static_cast<check_t>(&checks::chk_clear)}},
    {"collect", {
        static_cast<method_t>(&methods::scr_collect),
        static_cast<check_t>(&checks::chk_collect)}},
    {"answer", {
        static_cast<method_t>(&methods::scr_answer),
        static_cast<check_t>(&checks::chk_keyword)}},
    {"input", {
        static_cast<method_t>(&methods::scr_input),
        static_cast<check_t>(&checks::chk_collect)}},
    {"keystore", {
        static_cast<method_t>(&methods::scr_keystore),
        static_cast<check_t>(&checks::chk_keystore)}},
    {"store", {
        static_cast<method_t>(&methods::scr_store),
        static_cast<check_t>(&checks::chk_store)}},
};

Driver::Driver(const std::string& id) noexcept
{
    assert(driver_ == nullptr);
    driver_ = this;
    name_ = id;
    limit_ = 0;
}

Driver::~Driver() = default;

void Driver::configure(keyfile& config) // NOLINT
{
    if(!limit_) {
        const auto& server = config.at("server");
        try {
            limit_ = std::stoi(server.at("limit"));
        }
        catch(...) {
            limit_ = 20;
        }
    }

    auto size = sizeof(ipc_system) + limit_ * sizeof(ipc_session);
    auto shm = shm_open(BORDEAUX_IPCPATH, O_RDWR | O_CREAT, 0660);
    if(shm < 0 || ftruncate(shm, static_cast<off_t>(size)) < 0)
        log_fail(97, "failed to allocate shared memory");
    auto mem = mmap(nullptr, size, PROT_WRITE, MAP_SHARED, shm, 0);
    if(mem == MAP_FAILED)    // NOLINT
        log_fail(97, "failled to map shared memory");

    memset(mem, 0, size);
    sys_mapped = static_cast<ipc_system*>(mem);
    ipc_mapped = reinterpret_cast<ipc_session*>(&sys_mapped[1]);

    for(auto idx = 0U; idx < limit_; ++idx)
        ipc_free.push_back(&ipc_mapped[idx]);

    time(&sys_mapped->started);
    sys_mapped->level = logging;
    sys_mapped->limit = limit_;
    sys_mapped->active = false;

    timers = config.at("timers");
    try {
        ring_timer_ = std::stoi(timers.at("ring"));
    }
    catch(...) {}
    try {
        answer_timer_ = std::stoi(timers.at("answer"));
    }
    catch(...) {}
}

void Driver::restart([[maybe_unused]] const keyfile& config)
{
    std::lock_guard<std::mutex> lock(script_lock);
    server::dir dir(BORDEAUX_SCRIPTS_PATH);
    for(auto filename = "*"s; !filename.empty(); filename = dir.get()) {
        if(ends_with(filename, SCRIPT_EXT) && filename[0] != '_')
            scripts[filename] = script(BORDEAUX_SCRIPTS_PATH "/" + filename);
    };
}

void Driver::shutdown()
{
}

void Driver::disconnect()
{
    std::lock_guard<std::mutex> lock(session_lock);
    for(auto const& [sid, session] : sessions) {
        Session::send(session, shutdown_event);
    }
    sessions.clear();
}

void Driver::drop(Session *session)
{
}

void Driver::answer(Session *session)
{
}

auto Driver::config_instance(keyfile& cfg) -> keyfile
{
    assert(driver_ != nullptr);
    driver_->configure(cfg);
    return cfg;
}

void Driver::start_instance(const keyfile& ini)
{
    assert(driver_ != nullptr);
    log_info("starting {} driver with {} sessions", driver_->name(), driver_->limit_);
    driver_->startup(ini);
}

void Driver::reload_instance(const keyfile& cfg)
{
    assert(driver_ != nullptr);
    driver_->restart(cfg);
}

void Driver::stop_instance()
{
    assert(driver_ != nullptr);
    log_info("stopping {} driver", driver_->name());
    driver_->shutdown();
}

void Driver::drop_instance()
{
    assert(driver_ != nullptr);
    log_info("disconnect {} driver", driver_->name());
    driver_->disconnect();
}

auto Driver::attach(session_t& s) -> bool
{
    debug(3, "sid({}): attaching to driver", s->sid());
    std::lock_guard<std::mutex> lock(session_lock);
    if(ipc_free.empty())
        return false;
    sessions[s->sid()] = s;
    s->shm = ipc_free.front();
    s->slot_ = s->shm - ipc_mapped;
    time(&s->shm->created);
    util::set_string(s->shm->sid, 10, s->sid());
    s->shm->state = ipc_session::UNKNOWN;
    s->shm->type = ipc_session::ANY;
    ipc_free.pop_front();
    ++sys_mapped->used;
    return true;
}

void Driver::release(std::string_view sid)
{
    debug(3, "sid({}): releasing from driver", sid);
    Session::sync_store();
    std::lock_guard<std::mutex> lock(session_lock);
    try {
        auto& session = sessions.at(sid);
        driver_->drop(session.get());
        if(session->shm) {
            memset(session->shm, 0, sizeof(ipc_session));
            ipc_free.push_back(session->shm);
            --sys_mapped->used;
            session->shm = nullptr;
        }
        sessions.erase(sid);
    }
    catch(...) {
        log_warn("release: {}: unknown session", sid);
    }
}

auto Driver::control(const std::vector<std::string_view>& list) -> bool
{
    if(list.size() < 2)
        return false;

    try {
        auto s = get(list[1]);
    }
    catch(...) {
        return false;
    }

    return true;
}

auto Driver::timer(const char *id, int value) -> int
{
    try {
        return std::stoi(timers.at(id));
    }
    catch(...) {
        return value;
    }
}

Plugin::Plugin(std::string id) noexcept : name_(std::move(id)), next_(plugins_)
{
    plugins_ = this;
}

Plugin::~Plugin() = default;

void Plugin::startup([[maybe_unused]]const keyfile& config) {}
void Plugin::shutdown() {}
void Plugin::restart([[maybe_unused]]const keyfile& config) {}
void Plugin::configure([[maybe_unused]]keyfile& config) {}
auto Plugin::format([[maybe_unused]]Session *s, [[maybe_unused]]const char *keyword) -> const char * {return nullptr;}

void Plugin::config(keyfile& cfg)
{
    auto plugin = Plugin::plugins_;
    while(plugin) {
        plugin->configure(cfg);
        plugin = plugin->next_;
    }
}

void Plugin::start(const keyfile& cfg)
{
    auto plugin = Plugin::plugins_;
    while(plugin) {
        plugin->startup(cfg);
        plugin = plugin->next_;
    }
}

void Plugin::reload(const keyfile& cfg)
{
    auto plugin = Plugin::plugins_;
    while(plugin) {
        plugin->restart(cfg);
        plugin = plugin->next_;
    }
}

void Plugin::stop()
{
    auto plugin = Plugin::plugins_;
    while(plugin) {
        plugin->shutdown();
        plugin = plugin->next_;
    }
}

auto Plugin::find_locale(std::string_view from) -> Plugin *
{
    auto plugin = Plugin::plugins_;
    while(plugin) {
        if(plugin->name() == from)
            return plugin;
        plugin = plugin->next_;
    }

    auto pos = from.find('_');
    if(pos == std::string::npos || pos < 2)
        return nullptr;

    from.remove_suffix(from.size() - pos);
    plugin = Plugin::plugins_;
    while(plugin) {
        if(plugin->name() == from)
            return plugin;
        plugin = plugin->next_;
    }
    return nullptr;
}

auto Plugin::get_formatted(Session *s, const char *keyword) -> const char *
{
    auto plugin = Plugin::plugins_;
    while(plugin) {
        if(plugin->name() == keyword)
            return plugin->format(s, keyword);
        plugin = plugin->next_;
    }
    return nullptr;
}

