/*
 * Copyright (C) 2017-2021 David Sugar <tychosoft@gmail.com>.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "media.hpp"
#include "drivers.hpp"
#include "sessions.hpp"
#include "logger.hpp"
#include "proc.hpp"

#include <strings.hpp>

#include <poll.h>

using namespace server;

rtp_audio::rtp_audio(Session *s, std::string local_media, std::string_view& sdp, uint16_t port) :   // NOLINT
media(std::move(local_media)), session_(s), remote_to{}, remote_from{}, input_buffer{}, output_buffer{}, input_packet(input_buffer), output_packet(output_buffer)
{
    sent_total = recv_total = skip_total = fail_total = 0;

    ignore();
    std::tie(socket_, local_port) = net::bind("*", port);
    if(!local_port)
        log_error("failed to bind media to {}", media);
    else
        debug(4, "rtp media mapped to {}:{}", media, local_port);

    apply_sdp(sdp);
    debug(4, "media connected to {}:{}", remote_media, remote_port);
    output_packet.set_payload(payload, samples_, framing_);
    output_packet.set_version();
    output_packet.set_source(port ^ getpid());
}

rtp_audio::~rtp_audio()
{
    try {
        finish();
    } catch(...) {};
}

void rtp_audio::apply_sdp(std::string_view& sdp)
{
    if(sdp.empty())
        return;

    dtmf = payload = 0;     // clear before changing
    auto lines =tokenize(sdp, "\r\n");
    for(auto line : lines) {
        if(line.size() < 16)
            continue;

        if(util::compare(line, "c=IN IP4 ") && !ipv6) {
            line.remove_prefix(9);
            remote_media = std::string{line};
            continue;
        }

        if(util::compare(line, "c=IN IP6 ") && ipv6) {
            line.remove_prefix(9);
            remote_media = std::string{line};
            continue;
        }

        if(util::compare(line, "m=audio ")) {
            auto items = tokenize(line, " ");
            if(items.size() < 4)
                continue;
            remote_port = util::get_value(items[1]);
            continue;
        }

        if(!util::compare(line, "a=rtpmap:"))
            continue;

        line.remove_prefix(9);
        auto items = tokenize(line, " ");
        if(items.size() != 2)
            continue;
        if(util::compare(items[1], "pcmu/8000"))
            payload = util::get_value(items[0]);
        else if(util::compare(items[1], "telephone-event/8000"))
            dtmf = util::get_value(items[0]);
    }
    net::store(remote_to, remote_media, remote_port);
}

void rtp_audio::start()
{
    if(active_)
        return;

    silent();
    timer_ = util::stable_clock() + std::chrono::milliseconds(framing_);
    active_ = true;
    thread_ = std::thread(&rtp_audio::exec_thread, this);
}

void rtp_audio::finish()
{
    if(active_) {
        active_ = false;
        thread_.join();
    }
    audio_.close();
    if(socket_ > -1) {
        ::close(socket_);
        socket_ = -1;
        debug(4, "rtp media released from {}:{}", media, local_port);
    }
}

auto rtp_audio::sync_timer() noexcept -> milliseconds_t
{
    for(;;) {
        auto now = util::stable_clock();
        if(now < timer_)
            return std::chrono::duration_cast<std::chrono::milliseconds>(timer_ - now).count() + 1;
        timer_ += std::chrono::milliseconds(framing_);
        if(now < timer_)
            return 0;
        ++skip_total;
        output_packet.inc_silence();
    }
}

void rtp_audio::exec_thread()
{
    struct pollfd poller{};
    poller.fd = socket_;
    poller.events = POLLIN;

    debug(3, "sid({}): starting media thread", session_->sid());
    proc::set_priority(2);
    while(running && active_) {
        auto timeout = sync_timer();
        if(!timeout) {
            output();
            continue;
        }
        poller.revents = 0;
        if(poll(&poller, 1, static_cast<int>(timeout)) == 1 && (poller.revents & POLLIN))
            input();
    }
    debug(2, "sid({}): sent={}, recv={}, skips={}, failed={}", session_->sid(), sent_total, recv_total, skip_total, fail_total);
}

void rtp_audio::keepalive()
{
    std::lock_guard<std::mutex> lock(mutex_);
    if(address()->sa_family != AF_UNSPEC)
        ::sendto(socket_, nullptr, 0, 0, address(), net::size(address()));
}

auto rtp_audio::output() noexcept -> size_t
{
    std::lock_guard<std::mutex> lock(mutex_);
    if(address()->sa_family == AF_UNSPEC)
        return 0;

    auto size = (this->*(put_))(output_packet);
    if(size > 0) {
        if(output_packet.send(socket_, address(), size) < 0) {
            ++fail_total;
            if(errno == EAGAIN) {     // blocked we skip as if silent...
                output_packet.inc_silence();
                ++skip_total;
                return size;
            }
            return 0;
        }
        ++sent_total;
        output_packet.inc_sequence();
        return size;
    }
    ++skip_total;
    output_packet.inc_silence();
    return 0;
}

auto rtp_audio::input() noexcept -> size_t
{
    auto size = input_packet.recv(socket_, from(), max_audio_size);
    if(size < 0)
        return 0;
    // if invalid packet, then we ignore...
    if(!sync_address) {
        auto alen = net::size(from());
        size = memcmp(from(), address(), alen) == 0 ? size : 0;
    }
    if(size) {
        if(input_packet.version() != 2)
            return 0;
        if(input_packet.payload() != payload && input_packet.payload() != dtmf)
            return 0;
    }
    // get offset if not set...
    if(!offset_)
        offset_ = input_packet.timestamp();

    // loose first packet if starts at timestamp == 0
    if(!offset_)
        return 0;

    input_packet.adjust_offset(offset_);
    ++recv_total;

    // special throw-away case, time reversal prior to offset...
    if(input_packet.timestamp() >= 0xf0000000)
        return 0;

    // see if we have audio...
    if(input_packet.payload() == payload) {
        std::lock_guard<std::mutex> lock(mutex_);
        return (this->*(get_))(input_packet);
    }
    // TODO: dtmf support...
    return 0;
}

void rtp_audio::ignore()
{
    std::lock_guard<std::mutex> lock(mutex_);
    counted_ = 0;
    offset_ = 0;
    audio_.close();
    get_ = &rtp_audio::get_ignore;
    put_ = &rtp_audio::put_ignore;
}

auto rtp_audio::silent() -> unsigned
{
    std::lock_guard<std::mutex> lock(mutex_);
    audio_.close();
    auto dp = output_packet.data();
    memset(dp, filler_, static_cast<size_t>(samples_) * bytes_);
    output_packet.set_payload(payload, samples_, framing_);
    get_ = &rtp_audio::get_ignore;
    put_ = &rtp_audio::put_silent;
    auto timed = (counted_ / samples_) / fps_;
    counted_ = 0;
    offset_ = 0;
    return timed;
}

auto rtp_audio::put_silent([[maybe_unused]] rtp_packet& rtp) noexcept -> size_t // NOLINT
{
    return size_t(samples_);
}

auto rtp_audio::put_ignore([[maybe_unused]] rtp_packet& rtp) noexcept -> size_t // NOLINT
{
    return size_t(0);
}

auto rtp_audio::get_ignore([[maybe_unused]] rtp_packet& rtp) noexcept -> size_t // NOLINT
{
    return size_t(0);
}
