/*
 * Copyright (C) 2020-2021 David Sugar <tychosoft@gmail.com>.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "drivers.hpp"
#include "runtime.hpp"
#include "logger.hpp"

#include <strings.hpp>

using namespace server;

auto methods::scr_pause() noexcept -> bool
{
    clear_timeout();
    set_handler("pause", &methods::sleep_handler);
    return false;
}

auto methods::scr_exit() noexcept -> bool
{
    line = nullptr;
    return true;
}

auto methods::scr_switch() noexcept -> bool
{
    if(index >= line->argc) {
        skip();
        return true;
    }

    if(conditional()) {
        skip();
        return false;
    }

    return goto_end();
}

auto methods::scr_case() noexcept -> bool
{
    if(conditional()) {
        while(line && eq(line->cmd, "case"))    // cascade fallthrue...
            skip();
        return false;
    }

    index = 0;

    auto top = line->top;
    while(line) {
        line = line->next;
        if(line->top == top && eq(line->cmd, "otherwise")) {
            skip();
            return false;
        }
        if(line->top == top && eq(line->cmd, "end"))
            return false;
    }

    return false;;
}

auto methods::scr_otherwise() noexcept -> bool
{
    return goto_end();
}

auto methods::scr_break() noexcept -> bool
{
    if(!conditional()) {
        skip();
        return false;
    }
    goto_end();
    return false;
}

auto methods::scr_end() noexcept -> bool
{
    if(index >= line->argc) {
        skip();
        return true;
    }

    return scr_repeat();
}

auto methods::scr_repeat() noexcept -> bool
{
    if(!conditional()) {
        skip();
        return false;
    }

    if(line->top) {
        line = line->top;
        index = 0;
    }
    else
        entry();
    return false;
}

auto methods::scr_event() noexcept -> bool
{
    while(auto signal = value()) {
        if(event(signal))
            return false;
    }
    return error("trap not found");
}

auto methods::scr_do() noexcept -> bool
{
    while(auto next = value()) {
        if(label(next))
            return false;
    }
    return error("label not found");
}

auto methods::scr_set() noexcept -> bool
{
    auto sym = symbol();
    if(!sym)
        return error("var missing");

    char buffer[symbols::var_size + 1];
    buffer[0] = 0;
    while(auto v = value())
        mem_append(buffer, sizeof(buffer), v);

    if(!syms.set_value(sym, buffer))
        return error("cannot set value");

    skip();
    return true;
}

auto methods::scr_store() noexcept -> bool
{
    try {
        auto store = syms.get_value("KEYSTORE");
        if(!store || !*store) {
            if(getppid() == 1)
                log_error(":sid({}): cannot get keystore", sid());
            return error("cannot get keystore");
        }
        put_store(store);
    }
    catch(...) {}
    skip();
    return true;
}

auto methods::scr_keystore() noexcept -> bool
{
    try {
        auto store = value();
        if(!store || !*store) {
            if(getppid() == 1)
                log_error(":sid({}): cannot get keystore", sid());
            return error("cannot get keystore");
        }
        syms.make_const("KEYSTORE", store);
        get_store(store);
    } catch(...) {}
    skip();
    return true;
}

auto methods::scr_debug() noexcept -> bool
{
    auto delim=" ";
    char buffer[256];
    try {
        while(auto opt = option()) {
            auto val = value();
            if(!val)
                return error("missing value");

            if(eq(opt, "delim"))
                delim = val;
            else
                return error("invalid option");
        }

        auto level = value()[0] - '0';
        buffer[0] = 0;
        while(auto v = value()) {
            if(buffer[0] && delim && delim[0])
                mem_append(buffer, sizeof(buffer), delim);
            mem_append(buffer, sizeof(buffer), v);
        }
        debug(level, "sid({}): {}", sid(), buffer);
    } catch(...) {}
    skip();
    return true;
}

auto methods::scr_const() noexcept -> bool
{
    if(!eq(header->id, "_init_"))
        return error("can only be initializer");

    auto sym = symbol();
    if(!sym)
        return error("const missing");

    char buffer[symbols::var_size + 1];
    buffer[0] = 0;
    while(auto v = value())
        mem_append(buffer, sizeof(buffer), v);

    // make sure defined once only...
    if(!syms.exists(sym))
        syms.make_const(sym, buffer);
    skip();
    return true;
}

auto methods::scr_clear() noexcept -> bool
{
    if(line->argc == 0)
        digits[0] = 0;
    else while(auto sym = symbol())
        syms.clear_symbol(sym);

    if(value())
        return error("invalid arguments");

    skip();
    return true;
}

auto methods::scr_input() noexcept -> bool
{
    digits[0] = 0;
    return scr_collect();
}

auto methods::scr_collect() noexcept -> bool
{
    state.collect.max_digits = 3;
    state.collect.min_digits = 1;
    state.collect.wait_until = 10000L;
    state.collect.interdigit = 1000L;
    state.collect.digits = digits;

    auto max_size = max_digits;

    while(auto opt = option()) {
        auto val = value();
        if(!val)
            return error("missing value");

        if(eq(opt, "interdigit"))
            state.collect.interdigit = util::to_duration(val);
        else if(eq(opt, "timeout"))
            state.collect.wait_until = util::to_duration(val);
        else if(eq(opt, "digits"))
            state.collect.max_digits = static_cast<size_t>(symbols::to_value(val));
        else if(eq(opt, "requires"))
            state.collect.min_digits = static_cast<size_t>(symbols::to_value(val));
        else
            return error("invalid option");
    }

    auto sym = symbol();
    if(sym) {
        state.collect.digits = syms.make_storage(sym, static_cast<unsigned short>(state.collect.max_digits));
        if(!state.collect.digits) {
            state.collect.digits = syms.get_storage(sym);
            if(!state.collect.digits)
                return error("symbol not defined");
        }
        max_size = syms.get_size(sym);
        syms.set_value(sym, digits);       // copy and clear digits buffer
        digits[0] = 0;
    }

    if(state.collect.max_digits > max_size)
        return error("too many digits");

    if(state.collect.min_digits > state.collect.max_digits)
        return error("too many required");

    if(!state.collect.max_digits || !state.collect.min_digits || !state.collect.interdigit || !state.collect.wait_until)
        return error("invalid value");

    set_handler("collect", &methods::collect_handler);
    return true;
}

auto methods::scr_var() noexcept -> bool
{
    auto size = symbols::var_size;

    while(auto opt = option()) {
        auto val = value();
        if(!val)
            return error("missing value");
        if(eq(opt, "size")) {
          size = symbols::to_size(val);
          if (size < 1 || size > symbols::max_array)
            return error("invalid size");
        }
        else
            return error("invalid option");
    }

    while(auto sym = symbol())
        syms.make_value(sym, size);

    if(value())
        return error("invalid arguments");

    skip();
    return true;
}

auto methods::scr_answer() noexcept -> bool
{
    if(connected) {
        skip();
        return true;
    }

    Driver::instance()->answer(this);
    auto until = util::to_duration("30s");
    set_timeout(timer(until));
    set_handler("answer", &methods::answer_handler);
    return false;
}

auto methods::scr_sleep() noexcept -> bool
{
    const char *cp = value();
    if(!cp)
        return error("missing value");

    auto until = util::to_duration(cp);
    if(until < 1) {
        skip();
        return true;
    }

    set_timeout(timer(until));
    set_handler("sleep", &methods::sleep_handler);
    return false;
}

auto methods::scr_until() noexcept -> bool
{
    const char *cp = value();
    if(!cp)
        return error("missing value");

    auto until = util::to_duration(cp);
    if(until < 1) {
        skip();
        return true;
    }
    auto wait = started;
    wait += until;
    if(wait.until() < 50)
        return error("time expired");

    set_timeout(wait);
    set_handler("sleep", &methods::sleep_handler);
    return false;
}

auto checks::chk_keyword(const script::context_t& ctx) -> const char *
{
    if(ctx.header && eq(ctx.header->id, "_init_"))
        return "cannot be used to initialize";

    if(ctx.line && ctx.line->argc)
        return "arguments not used";
    return nullptr;
}

auto checks::chk_conditional(const script::context_t& ctx) -> const char *
{
    if(ctx.header && eq(ctx.header->id, "_init_"))
        return "cannot be used to initialize";

    if(!ctx.line || !ctx.line->argc)
        return nullptr;

    // TODO: expression verification... chk_expression
    return nullptr;
}

auto checks::chk_otherwise(const script::context_t& ctx) -> const char *
{
    if(ctx.header && eq(ctx.header->id, "_init_"))
        return "cannot be used to initialize";

    if(ctx.line && ctx.line->argc)
        return "arguments not used";

    if(ctx.line && !ctx.line->top)
        return "can only be in a block";

    // TODO: expression verification... chk_expression
    return nullptr;
}

auto checks::chk_case(const script::context_t& ctx) -> const char *
{
    if(ctx.header && eq(ctx.header->id, "_init_"))
        return "cannot be used to initialize";

    if(!ctx.line || !ctx.line->argc)
        return "requires conditional arguments";

    auto top = ctx.line->top;
    if(!top)
        return "can only be in a block";

    top = top->next;
    if(top && !eq(top->cmd, "case"))
        return "can only be in cascading case block";

    // TODO: expression verification... chk_expression
    return nullptr;
}

auto checks::chk_const(const script::context_t& ctx) -> const char *
{
    if(ctx.header && !eq(ctx.header->id, "_init_"))
        return "can only be initialize";

    if(!ctx.line || !ctx.line->argc)
        return "const symbol required";

    auto cp = ctx.line->argv[0];
    if(*cp == '%')
        ++cp;

    if(!script::is_constant(cp))
        return "consts must be upper case";

     return nullptr;
}

auto checks::chk_clear(const script::context_t& ctx) -> const char *
{
    if(!ctx.line || !ctx.line->argc)
        return nullptr; // by default clears %digits...

    unsigned index = 0;
    const char *cp = nullptr;
    while(nullptr != (cp = ctx.line->argv[index])) {
        if(*cp == '%')
            ++cp;

        if(!script::is_variable(cp))
            return "vars must be lower case and have single scope";

        ++index;
    }
    return nullptr;
}

auto checks::chk_var(const script::context_t& ctx) -> const char *
{
    if(!ctx.line || !ctx.line->argc)
        return "var symbol required";

    unsigned index = 0;
    const char *cp = nullptr;
    while(nullptr != (cp = ctx.line->argv[index])) {
        if(*cp != '=')
            break;
        cp = ctx.line->argv[++index];
        if(!cp)
            return "var option missing";
        ++index;
    }

    while(nullptr != (cp = ctx.line->argv[index])) {
        if(*cp == '%')
            ++cp;

        if(!script::is_variable(cp))
            return "vars must be lower case and have single scope";
        ++index;
    }
    return nullptr;
}

auto checks::chk_collect(const script::context_t& ctx) -> const char *
{
    if(ctx.header && eq(ctx.header->id, "_init_"))
        return "cannot be used to initialize";

    unsigned index = 0;
    const char *cp = nullptr;
    while(nullptr != (cp = ctx.line->argv[index])) {
        if(*cp != '=')
            break;
        cp = ctx.line->argv[++index];
        if(!cp)
            return "var option missing";
        ++index;
    }

    if(index < ctx.line->argc - 1)
        return "only one var allowed";

    cp = ctx.line->argv[index];
    if(!cp)
        return nullptr;

    if(*cp == '%')
        ++cp;

    if(!script::is_variable(cp))
        return "vars must be lower case and have single scope";

     return nullptr;
}

auto checks::chk_keystore(const script::context_t& ctx) -> const char *
{
    if(ctx.header && !eq(ctx.header->id, "_init_"))
        return "can only be initializer";

    if(ctx.level > 0)
        return "cannot be imported";

    if(!ctx.line || ctx.line->argc != 1)
        return "can only have one argument";

    if(ctx.keystore)
        return "keystore already defined";

    ctx.keystore = true;
    return nullptr;
}

auto checks::chk_debug(const script::context_t& ctx) -> const char *
{
    if(!ctx.line || !ctx.line->argc)
        return "debug level required";

    unsigned index = 0;
    const char *cp = nullptr;
    while(nullptr != (cp = ctx.line->argv[index])) {
        if(*cp != '=')
            break;
        cp = ctx.line->argv[++index];
        if(!cp)
            return "var option missing";
        ++index;
    }

    if(!ctx.line || index >= ctx.line->argc - 1 || !ctx.line->argv[index])
        return "at least one debug message required";

    auto sv = std::string_view(ctx.line->argv[index]);
    if(sv < "#0" || sv > "#9")
        return "debug level must be 0-9";

     return nullptr;
}

auto checks::chk_store(const script::context_t& ctx) -> const char *
{
    if(!ctx.line || !ctx.line->argc)
        return "symbols required";

    auto index = 0U;
    auto cp = "";
    while(nullptr != (cp = ctx.line->argv[index++])) {
        if(*cp == '%')
            ++cp;
        if(!script::is_variable(cp))
            return "vars must be lower case and have single scope";
    }

    return nullptr;
}

auto checks::chk_set(const script::context_t& ctx) -> const char *
{
    if(!ctx.line || !ctx.line->argc)
        return "set symbol required";

    auto cp = ctx.line->argv[0];
    if(*cp == '%')
        ++cp;

    if(!script::is_variable(cp))
        return "vars must be lower case and have single scope";

     return nullptr;
}

auto checks::chk_do(const script::context_t& ctx) -> const char *
{
    if(ctx.header && eq(ctx.header->id, "_init_"))
        return "cannot be used to initialize";

    if(ctx.line && !ctx.line->argc)
        return "at least one label required";

    return nullptr;
}

auto checks::chk_sleep(const script::context_t& ctx) -> const char *
{
    if(ctx.header && eq(ctx.header->id, "_init_"))
        return "cannot be used to initialize";

    if(!ctx.line || ctx.line->argc == 0)
        return "no duration specified";

    if(ctx.line->argc > 1)
        return "too many arguments";

    return nullptr;
}

auto checks::chk_event(const script::context_t& ctx) -> const char *
{
    if(ctx.header && eq(ctx.header->id, "_init_"))
        return "cannot be used to initialize";

    if(ctx.line && !ctx.line->argc)
        return "at least one event required";

    return nullptr;
}
