/* Copyright (C) 2020-2021 David Sugar <tychosoft@gmail.com>.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "drivers.hpp"
#include "media.hpp"
#include "sessions.hpp"
#include "logger.hpp"
#include "proc.hpp"

#include <strings.hpp>
#include <keyfile.hpp>

#include <atomic>
#include <mutex>

using namespace server;

namespace {
std::atomic<sid_t> next_sid{1};
Session::runlist_t runlist;
std::mutex runlock;
pthread_condattr_t cattr;
pthread_cond_t cond;
keyfile key_storage;
std::mutex storelock;
std::atomic<bool> key_updates{false};

const Session::event_t entry_event{event_type::entry};
const Session::event_t leave_event{event_type::leave};
const Session::event_t timeout_event{event_type::timeout};
const Session::event_t exit_event{event_type::exit};


} // end anon namespace

Session::Session(const script& attach) noexcept :
language(nullptr), handler(&Session::dummy_handler), expires(started), image(attach), shm(nullptr), audio(nullptr), index(0), level(0), connected(false), exiting(false), media(false), run_(runlist.cend()) {
    auto seq = next_sid.fetch_add(1, std::memory_order_relaxed);
    uint8_t sid64[6];

    util::random_fill(sid64, 4);
    sid64[4] = seq & 0xff;
    sid64[5] = seq & 0xff00 >> 8;

    var_lang = syms.make_storage("LANGUAGE", 16, "en_US");  // NOLINT
    var_voice = syms.make_storage("VOICE", 64, util::get_env("VOICE", 64)); // NOLINT
    var_mailbox = syms.make_storage("MAILBOX", 32, "none"); // NOLINT

    set_lang(util::get_env("LANGUAGE", 16));

    syms.make_storage("result_code", 7);
    syms.set_value("result_code", "0");

    syms.make_const({
        {"SESSION_ID", util::to_b64(sid64, sizeof(sid64))},
        {"SCRIPT_NAME", attach.name()},
    });

    app_ = syms.get_value("SCRIPT_NAME");   // NOLINT
    sid_ = syms.get_value("SESSION_ID");    // NOLINT
    mem_copy(stats.sid, sizeof(stats.sid), sid_);
    mem_copy(stats.state, sizeof(stats.state), "init");
    stats.lines = stats.steps = 0;
    stats.start = util::current_time();

    var_audio = static_cast<audio_paths*>(syms.alloc(sizeof(audio_paths))); // NOLINT
    var_audio->app = app_;
    var_audio->lang = var_lang;
    var_audio->mailbox = var_mailbox;
    var_audio->voice = var_voice;
}

Session::~Session()
{
    disconnect();
    mem_copy(stats.state, sizeof(stats.state), "down");
    memset(stats.sid, 0, sizeof(stats.sid));

    // clean up any session recording
    auto filename = "temp/"s + sid_ + ".au";
    ::remove(filename.c_str());
}

void Session::begin(const session_t& s, const char *name)
{
    s->begin(name);
    send(s, entry_event);
}

void Session::begin(const char *name)
{
    error_ = syms.make_storage("error", symbols::var_size);
    timed_ = syms.make_storage("timed", 16);    // timed media
    digits = syms.make_storage("digits", max_digits);
    index = level = 0;
    items = nullptr;

    // run init segment...
    header = image.init();
    line = header->first;
    while(line)
        execute_method(line->method);

    label(name);
    set_timeout(step_rate);
    set_handler("step", &Session::script_handler);
}

auto Session::set_handler(const char *text, handler_t new_handler) -> bool
{
    if(handler == new_handler)
        return false;

    process_event(leave_event);
    handler = new_handler;
    mem_copy(stats.state, sizeof(stats.state), text);
#ifdef  DEBUG
    debug(4, "sid({}): state={}", sid_, text);
#endif
    silence();
    return process_event(entry_event);
}

void Session::stream()
{
    if(!audio || connected)
        return;

    shm->state = ipc_session::CONNECTED;
    connected = true;
    audio->start();
}

void Session::silence()
{
    if(!audio || !media)
        return;

    media = false;
    set_timed(audio->silent());
}

void Session::connect(std::string local_media, std::string_view& sdp, uint16_t port = 0)
{
    disconnect();
    audio = new rtp_audio(this, local_media, sdp, port);    // NOLINT
}

void Session::libexit()
{
    if(fd_ != -1)
        ::close(fd_);

    if(pid_ != -1)
        proc::stop(pid_);

    fd_ = -1;
    pid_ = -1;
}

void Session::disconnect()
{
    libexit();
    if(audio) {
        audio->finish();
        delete audio;
        audio = nullptr;
    }
}

void Session::apply_sdp(std::string_view& sdp)
{
    if(audio)
        audio->apply_sdp(sdp);
}

void Session::clear_timeout()
{
    expires = started;
}

void Session::set_timeout(const timer& until)
{
    expires = until;
}

void Session::set_timeout(milliseconds_t offset)
{
    expires = timer(offset);
}

void Session::bump_timeout(milliseconds_t offset)
{
    expires += offset;
}

void Session::sync_store()
{
    auto flag = key_updates.exchange(false);
    if(!flag)
        return;

    std::lock_guard<std::mutex> lock(storelock);
    key_storage.write("stored.keys");
    key_updates = false;
}

void Session::set_lang(const char* from)
{
    if(!from || *from == 0)
        util::set_string(var_lang, 16, "en_US");
    else
        util::set_string(var_lang, 16, from);

    char *cp = strchr(var_lang, '.');
    if(cp && *cp == '.')
        *cp = 0;

    language = Plugin::find_locale(var_lang);
    cp = var_lang;
    while(*cp) {
        *cp = static_cast<char>(tolower(*cp));
        ++cp;
    }

    std::string path = BORDEAUX_VOICES_PATH "/";
    path += from;
    cp = strchr(var_lang, '_');
    if(util::is_dir(path) && cp  && *cp == '_')
        *cp = '-';
    else if(cp && *cp == '_')
        *cp = 0;
}

void Session::get_store(const char *id)
{
    std::lock_guard<std::mutex> lock(storelock);
    const auto& keys = key_storage[id];
    const auto& common = key_storage["_"];

    // load common default keys if any
    for(const auto& [key, value] : common) {
        snprintf(temp, sizeof(temp), "stored_%s", key.c_str());
        syms.set_value(temp, value.c_str());
    }

    // load target storage keys
    for(const auto& [key, value] : keys) {
        snprintf(temp, sizeof(temp), "stored_%s", key.c_str());
        syms.set_value(temp, value.c_str());
    }
}

void Session::put_store(const char *id)
{
    std::lock_guard<std::mutex> lock(storelock);
    auto& keys = key_storage[id];
    auto key = "";
    while(nullptr != (key = symbol())) {
        if(!strncmp(key, "stored_", 7))
            key += 7;
        snprintf(temp, sizeof(temp), "stored_%s", key);
        auto value = syms.get_value(temp);
        keys.insert_or_assign(key, value);
    }
    key_updates = true;
}

void Session::release(const session_t& s)
{
    if(s->expires == s->started && s->run_ == runlist.cend())
        return;

    std::unique_lock<std::mutex> lock(runlock);
    s->clear_timeout();
    if(s->run_ != runlist.cend())
        runlist.erase(s->run_);
    s->run_ = runlist.cend();

    lock.unlock();
    pthread_cond_signal(&cond);
}

void Session::start(int priority)
{
    key_storage.load("stored.keys");
    if(!proc::set_priority(priority))
        priority = 0;

    pthread_condattr_init(&cattr);
#ifndef __MACH__
    pthread_condattr_setclock(&cattr, CLOCK_MONOTONIC);
#endif
    pthread_cond_init(&cond, &cattr);

    log_info("starting session timers; pri={}", priority);
    while(running) {
        std::unique_lock<std::mutex> lock(runlock);
        if(runlist.empty()) {
#ifdef  DEBUG
            debug(2, "timers idle");
#endif
            pthread_cond_wait(&cond, runlock.native_handle());
        }
        else {
#ifdef  DEBUG
            debug(5, "timer triggered; runlist={}", runlist.size());
#endif
            timer now;
            auto s = *runlist.begin();
            auto expires = s->expires;
            auto mutex = runlock.native_handle();
            if(now < expires) {
#ifdef  DEBUG
                debug(5, "timers sleep {}", expires.until());
#endif
#ifdef  __MACH__
                until -= now;
                pthread_cond_timedwait_relative_np(&cond, mutex, &expires);
#else
                pthread_cond_timedwait(&cond, mutex, &expires);
#endif
            }
        }
        if(!running)
            break;

        timer now;
        while(!runlist.empty()) {
            // get earliest event remaining in list
            auto it = runlist.begin();
            auto s = *it;
            auto first = s->expires;

            // skip all future events after present...
            if(now < first)
                break;

            // timeout and remove this event
            runlist.erase(it);
            s->run_ = runlist.cend();
            lock.unlock();
            Session::send(s, timeout_event);
            lock.lock();
        }
    }
    log_info("stopping session timers");
    if(key_updates)
        key_storage.write("stored.keys");
}

void Session::stop()
{
    pthread_cond_signal(&cond);
    std::this_thread::sleep_for(std::chrono::milliseconds(20));
    std::lock_guard<std::mutex> lock(runlock);
    runlist.clear();
}

auto Session::send(const session_t& s, const event_t &event) -> bool
{
    std::unique_lock<std::mutex> session_lock(s->access_lock);
    auto result = s->process_event(event);
    if(s->started == s->expires && s->run_ == runlist.cend())
        return result;

    session_lock.unlock();
    std::unique_lock<std::mutex> runlist_lock(runlock);

    if(s->run_ != runlist.cend())
        runlist.erase(s->run_);

    if(s->started != s->expires) {
        auto [it, flag] = runlist.insert(s);
        s->run_ = it;
    }
    else
        s->run_ = runlist.cend();

    pthread_cond_signal(&cond);
    return result;
}

auto Session::common_handler(const event_t &event) noexcept -> bool
{
    size_t count = 0;
    switch(event.type) {
    case event_type::mediaerr:
        if(!media)
            return false;
        media = false;
        set_timed(event.timed);
        set_error("media");
        if(!Session::event("error"))
            skip();
        return this_step();
    case event_type::mediaend:
        if(!media)
            return false;
        media = false;
        set_timed(event.timed);
        if(!Session::event("media"))
            skip();
        return this_step();
    case event_type::yield:
        // if no active pid, then ignore...
        if(pid_ == -1)
            return false;

        // close off pid in session
        pid_ = -1;
        if(fd_ != -1)
            ::close(fd_);
        fd_ = -1;

        // yield pid result to scripting
        syms.set_value("result_code", std::to_string(event.result).c_str());
        return true;
    case event_type::drop:
        if(exiting) // can ignore call drops entirely if already exiting
            return true;
        // if hangup, send as sighup to libexec...
        if(pid_)
            proc::hangup(pid_);
        pid_ = -1;
    case event_type::exit:
        disconnect();
        shm->state = ipc_session::CLOSING;
        exiting = true;
        if(!trap && Session::event("exit"))
            return true;

        if(!eq(header->id, "exit") && label("exit"))
            return this_step();

        set_handler("idle", &Session::dummy_handler);
        clear_timeout();
        Driver::release(sid_);
        return false;
    case event_type::answer:
        stream();
        return true;
    case event_type::timeout:
        if(!Session::event("timeout"))
            skip();
        return this_step();
    case event_type::entry:
        return true;
    case event_type::flash:
        if(Session::event("flash") || label("flash"))
            return this_step();
        return true;
    case event_type::digit:
        if(Session::event(&event.digit[0]) || Session::event("digits")) {
            digits[0] = event.digit[0];
            digits[1] = 0;
            return this_step();
        }

        // trap handlers buffer for collect...
        count = mem_size(digits);
        if(count >= max_digits) {
            set_error("overflow");
            if(Session::event("overflow"))
                return this_step();
            return true;
        }
        digits[count++] = event.digit[0];
        digits[count] = 0;
        return false;
    case event_type::leave:
        libexit();
        return false;
    default:
        return false;
    }
}

auto Session::collect_handler(const event_t &event) noexcept -> bool
{
    auto store = state.collect.digits;
    auto count = mem_size(store);

    switch(event.type) {
    case event_type::digit:
        store[count++] = event.digit[0];
        store[count] = 0;
    case event_type::entry:
        if(count && strchr(state.collect.terminators, store[count - 1])) {
            store[--count] = 0;
            return next_step();
        }
        if(count >= state.collect.max_digits)
            return next_step();
        if(count < state.collect.min_digits)
            set_timeout(state.collect.wait_until);
        else
            set_timeout(state.collect.interdigit);
        return true;
    case event_type::leave:
        clear_timeout();
        return true;
    default:
        return common_handler(event);
    }
}

auto Session::answer_handler(const event_t& event) noexcept -> bool
{
    switch(event.type) {
    case event_type::timeout:
        return common_handler(exit_event);
    case event_type::answer:
        stream();
        return next_step();
    case event_type::drop:
    case event_type::exit:
        return common_handler(event);
    default:    // ignore most secondary events while waiting for answer
        return true;
    }
}

auto Session::sleep_handler(const event_t& event) noexcept -> bool
{
    switch(event.type) {
    case event_type::timeout:
        return next_step();
    default:
        return common_handler(event);
    }
}

auto Session::script_handler(const event_t& event) noexcept -> bool
{
    auto count = step_count;

    switch(event.type) {
    case event_type::entry:
    case event_type::timeout:
        if(!line)
            return common_handler(exit_event);

        set_timeout(step_rate);
        ++stats.steps;
        while(line && count--) {
            ++stats.lines;
            if(!execute_method(line->method))
                return false;
        }

        // if we really used up the entire interval, still sleep...
        if(!expires.until())
            set_timeout(step_rate);

        return true;
    case event_type::leave:
        return true;
    default:
        return common_handler(event);
    }
}

auto Session::dummy_handler(const event_t &event) noexcept -> bool
{
    switch(event.type) {
    case event_type::entry:
        clear_timeout();
        return true;
    default:
        return false;
    }
}

auto Session::argv(unsigned short offset) -> const char *
{
    if(offset >= line->argc)
        return nullptr;

    return line->argv[offset];
}

auto Session::format() -> const char *
{
    if(index >= line->argc)
        return nullptr;

    const char *cp = line->argv[index];
    if(*cp == '&')
        return ++cp;

    auto prior = index;
    auto older = items;
    cp = value();
    index = prior;
    items = older;

    if(!cp)
        return "invalid";
    if(!*cp)
        return "empty";

    auto len = mem_size(cp);
    auto dots = false;
    auto phone = false;

    if(len == 5 && cp[2] == ':' && symbols::to_value(cp, -1, 0, 23) != -1 && symbols::to_value(cp, -1, 0, 60) != -1)
        return "time";

    if(strchr(cp, ','))
        return "list";

    if(*cp == '-')
        ++cp;
    else if(*cp == '+') {
        phone = true;
        ++cp;
    }

    while(*cp) {
        if(*cp == '.') {
            if(!dots && !phone)
                dots = true;
            else
                break;
        }
        if(!isdigit(*cp) || (phone && isspace(*cp)))
            break;
        ++cp;
    }

    if(!*cp && dots)
        return "decimal";
    if(!*cp && phone)
        return "phone";
    if(!*cp)
        return "number";
    return "text";
}

auto Session::option() -> const char *
{
    if(index >= line->argc)
        return nullptr;

    auto cp = line->argv[index];
    if(*cp == '=')
        return ++cp;
    return nullptr;
}

auto Session::symbol() -> const char *
{
    while(index < line->argc) {
        auto cp = line->argv[index];
        if(*cp == '=') {
            ++index;
            continue;
        }
        if(*cp == '%') {
            ++index;
            return ++cp;
        }
        if(*cp == '$' || *cp == '#')
            return nullptr;
        ++index;
        return cp;
    }
    return nullptr;
}

auto Session::funcs(const char *cp) -> const char *
{
    const char *lp = nullptr;
    if(eq(cp, "&count")) {
        unsigned count = 0;
        const char *s = symbol();
        if(s) {
            snprintf(temp, sizeof(temp), "%u", syms.get_count(s));
            return temp;
        }

        const char *v = arg();
        if(!v)
            return "invalid";

        cp = v;
        while((cp = strchr(cp, ',')) != nullptr) {
            ++count;
            lp = ++cp;
        }

        if(lp && *lp)
            ++count;

        if(count) {
            mem_value(temp, sizeof(temp), count);
            return temp;
        }

        mem_value(temp, sizeof(temp), mem_size(v));
        return temp;
    }

    // nth value for ordering
    if(eq(cp, "&order")) {
        const char *v = arg();
        if(!v || *v == '0')
            return "invalid";
        cp = v;
        while(*cp && isdigit(*cp))
            ++cp;
        if(*cp)
            return "invalid";
        return v;
    }

    // simple number...
    if(eq(cp, "&number")) {
        const char *v = arg();
        if(!v)
            return "invalid";
        cp = v;
        if(*cp == '-')
            ++cp;
        while(*cp && isdigit(*cp))
            ++cp;
        if(*cp)
            return "invalid";
        return v;
    }

    // decimal number
    if(eq(cp, "&decimal")) {
        const char *v = arg();
        if(!v)
            return "invalid";
        bool dots = false;
        cp = v;
        if(*cp == '-')
            ++cp;
        while(*cp) {
            if(isdigit(*cp)) {
                ++cp;
                continue;
            }
            if(*cp == '.' && !dots) {
                dots = true;
                ++cp;
                continue;
            }
            return "invalid";
        }
        if(!dots) {
            snprintf(temp, sizeof(temp), "%s.", v);
            v = temp;
        }
        return v;
    }

    if(eq(cp, "&spell")) {
        const char *v = arg();
        if(!v || strchr(v, ',') || mem_size(v) > (sizeof(temp) / 2 - 2))
            return "invalid";
        char *tp = temp;
        while(*v) {
            *(tp++) = *(v++);
            *(tp++) = ',';
        }
        *tp = 0;
        return temp;
    }

    if(eq(cp, "&path")) {
        cp = arg();
        if(!cp)
            return "/nonexists";
        if(!audio_file::path(cp, temp, sizeof(temp), var_audio))
            return "/nonexists";
        return temp;
    }

    if(eq(cp, "&items")) {
        cp = arg();
        if(!cp)
            return "invalid";
        items = cp;
        cp = itemize();
        if(!cp)
            return "invalid";
        return cp;
    }

    if(eq(cp, "&list")) {
        const char *v = arg();
        if(!v)
            return ",";

        if(!strchr(v, ',')) {   // single element must be xxx,
            snprintf(temp, sizeof(temp), "%s,", v);
            v = temp;
            if(!strchr(v, ',')) // element too long...
                return "invalid";
        }
        return v;
    }

    if(eq(cp, "&time")) {
        const char *v = arg();
        if(!v)
            return "invalid";
        snprintf(temp, sizeof(temp), "%s", script::str_time24(v).c_str());
        return temp;
    }
    return "invalid";
}

auto Session::itemize() -> const char *
{
    if(items) {
        while(*items && !isalnum(*items) && *items != '/' && *items != '.' && *items != '-')
            ++items;
        if(*items) {
            auto offset = 0U;
            while((*items == '.' || *items == '/' || *items == '-' || isalnum(*items)) && offset < (sizeof(item) - 1))
                item[offset++] = *(items++);
            item[offset] = 0;
            return item;
        }
        items = nullptr;
    }
    return nullptr;
}

auto Session::value() -> const char *
{
    // skip empty items, itemize if from list
    auto cp = itemize();
    if(cp)
        return cp;

    // parse from argument
    while(index < line->argc) {
        cp = line->argv[index++];
        if(*cp == '=')
            continue;
        if(*cp == '$' || *cp == '#' || *cp == '@' || *cp == '^') // literal...
            return ++cp;
        if(*cp == '%')						// symbol...
            return syms.get_value(++cp);
        if(*cp == '&')
            return funcs(cp);
        return cp;
    }
    return nullptr;
}

auto Session::arg() -> const char *
{
    while(index < line->argc) {
        auto cp = line->argv[index++];
        if(*cp == '$' || *cp == '#' || *cp == '@' || *cp == '^') // literal...
            return ++cp;
        if(*cp == '%' || *cp == '&')						// symbol...
            return syms.get_value(++cp);
        return cp;
    }
    return nullptr;
}


auto Session::is_label(const char *id) -> bool
{
    assert(id != nullptr);

    if(*id == '@')
        ++id;

    auto section = image.init();
    while(section && !eq(id, section->id)) {
        section = section->next;
    }

    return section;
}

auto Session::goto_end() -> bool
{
    index = 0;
    items = nullptr;

    if(!line->top) {
        line = nullptr;
        return false;
    }

    auto top = line->top;
    while(line) {
        line = line->next;
        if(line->top == top && eq(line->cmd, "end"))
            return true;
    }

    return false;;
}

auto Session::conditional() -> bool
{
    if(index >= line->argc)
        return true;

    auto result = false;
    goto cont;

join:
    if(index >= line->argc)
        return result;

    if(eq(line->argv[index], "and") && !result)
        return false;

    if(!eq(line->argv[index], "or"))
        return false;

    ++index;
    goto cont;

cont:
    if(index >= line->argc)
        return result;

    auto cp = line->argv[index];
    if(eq(cp, "-defined")) {
        ++index;
        cp = value();
        result = (cp && *cp);
        goto join;
    }

    if(eq(cp, "-undefined")) {
        ++index;
        cp = value();
        result = !cp || !*cp;
        goto join;
    }

    auto v1 = value();
    auto op = line->argv[index++];
    auto v2 = value();

    if(!v1)
        v1 = "";

    if(!v2)
        v2 = "";

    if(!op)
        op = "?ne";

    if(eq(op, "?in")) {
        result = strstr(v2, v1) != nullptr;
        goto join;
    }

    if(eq(op, "?eq")) {
        result = eq(v1, v2);
        goto join;
    }

    if(eq(op, "?ne")) {
        result = !eq(v1, v2);
        goto join;
    }

    if(eq(op, "-eq")) {
        result = util::get_value(v1) == util::get_value(v2);
        goto join;
    }

    if(eq(op, "-ne")) {
        result = util::get_value(v1) != util::get_value(v2);
        goto join;
    }

    if(eq(op, "-gt")) {
        result = util::get_value(v1) > util::get_value(v2);
        goto join;
    }

    if(eq(op, "-ge")) {
        result = util::get_value(v1) >= util::get_value(v2);
        goto join;
    }

    if(eq(op, "-lt")) {
        result = util::get_value(v1) < util::get_value(v2);
        goto join;
    }

    if(eq(op, "-le")) {
        result = util::get_value(v1) <= util::get_value(v2);
        goto join;
    }

    result = false;
    goto join;
}

auto Session::label(const char *id) -> bool
{
    assert(id != nullptr);

    if(*id == '@')
        ++id;

    auto section = image.init();
    while(section && !eq(id, section->id)) {
        section = section->next;
    }

    if(!section)
        return false;

    header = section;
    entry();
    return true;
}

void Session::entry()
{
    assert(header != nullptr);

    set_error("none");
    line = header->first;
    trap = nullptr;
    digits[0] = 0;
    index = level = 0;
    items = nullptr;
}

auto Session::event(const char *id) -> bool
{
    assert(id != nullptr);

    if(*id == '^')
        ++id;

    if(eq(header->id, "_init_") || eq(header->id, "exit"))
        return false;

    auto jump = header->traps;
    while(jump && !eq(jump->id, id))
        jump = jump->next;

    if(jump && (!trap || jump->group != trap->group)) {
        trap = jump;
        line = trap->first;
        digits[0] = 0;
        if(!eq(jump->id, "error"))
            set_error("none");
        return true;
    }

    return false;
}

auto Session::error(const char *reason) -> bool
{
    if(getppid() > 1)
        print(std::cerr, "{}:{}:{}\n", image.name(), line->lnum, reason);

    set_error(reason);
    if(!eq(header->id, "_init_")) {
        if(event("error"))
            return false;
    }
    skip();
    return true;
}
