/*
 * Copyright (C) 2021 David Sugar <tychosoft@gmail.com>.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "utils.hpp"
#include "clock.hpp"

#include <print.hpp>

#include <fstream>
#include <iomanip>
#include <mutex>
#include <cstdio>

#include <unistd.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>

struct server::dir::stateinfo {
    DIR *handle;
    struct dirent *find;
};

using namespace server;

namespace {
std::ofstream logfile;
std::mutex loglock;

const char alphabet[65] =
    "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

auto u8_1b(const uint8_t *bp)
{
    return bp[0] == 0x09 || bp[0] == 0x0a || bp[0] == 0x0d || (bp[0] >= 0x20 && bp[0] <= 0x7e);
}

auto u8_2b(const uint8_t *bp)
{
     return bp[0] >= 0xc2 && bp[0] <= 0xdf && bp[1] >= 0x80 && bp[1] <= 0xbf;
}

auto u8_3b(const uint8_t *bp)
{
    if(bp[0] == 0xe0 && bp[1] >= 0xa0 && bp[1] <= 0xbf && bp[2] >= 0x80 && bp[2] <= 0xbf)
        return true;

    if(bp[0] >= 0xe1 && bp[0] <= 0xef && bp[0] != 0xed && bp[1] >=0x80 && bp[1] <= 0xbf && bp[2] >= 0x80 && bp[2] <= 0xbf)
        return true;

    if(bp[0] == 0xed && bp[1] >= 0x80 && bp[1] <= 0x9f && bp[2] >= 0x80 && bp[2] < 0xbf)
        return true;

    return false;
}

auto u8_4b(const uint8_t *bp)
{
    if(bp[0] == 0xf0 && bp[1] >= 0x90 && bp[1] <= 0xbf && bp[2] >= 0x80 && bp[2] <= 0xbf && bp[3] >= 0x80 && bp[3] <= 0xbf)
        return true;

    if(bp[0] >= 0xf1 && bp[0] <= 0xf3 && bp[1] >= 0x80 && bp[1] <= 0xbf && bp[2] >= 0x80 && bp[2] <= 0xbf && bp[3] >= 0x80 && bp[3] <= 0xbf)
        return true;

    if(bp[0] == 0xf4 && bp[1] >= 0x80 && bp[1] <= 0x8f && bp[2] >= 0x80 && bp[2] <= 0xbf && bp[3] >= 0x80 && bp[3] <= 0xbf)
        return true;

    return false;
}
}

void server::log_reset()
{
    const std::lock_guard<std::mutex> lock(loglock);
    if(logfile.is_open())
        logfile.close();
    logfile.open(SERVER_LOGGER_PATH "/bordeaux.log", std::ios_base::app); // FlawFinder: ignore
}

void server::logger(const std::string& msg)
{
    if(!logfile.is_open())
        return;

    time_t timestamp{};
    time(&timestamp);
    struct tm now = util::local_time(timestamp);
    const std::lock_guard<std::mutex> lock(loglock);
    logfile << std::put_time(&now, "%Y-%m-%d %X ") << msg << std::endl;
}

auto util::u8verify(const char *cp, size_t max) -> size_t
{
    if(!cp || !*cp || !max)
        return 0;

    size_t size = 0, codepoints = 0;
    auto bp = reinterpret_cast<const uint8_t *>(cp);
    while(*bp && codepoints < max) {
        if(u8_1b(bp)) {
            ++bp;
            ++size;
            ++codepoints;
            continue;
        }

        if(u8_2b(bp)) {
            size += 2;
            bp +=2;
            ++codepoints;
            continue;
        }

        if(u8_3b(bp)) {
            size += 3;
            bp += 3;
            ++codepoints;
            continue;
        }
        if(u8_4b(bp)) {
            size += 4;
            bp += 4;
            ++codepoints;
            continue;
        }

        return 0;
    }

    if(codepoints > max || *bp)
        return 0;

    return size;
}

auto util::compare(std::string_view from, std::string_view to) -> bool
{
    auto size = to.size();
    if(!size)
        return false;

    if(from.size() < size)
        return false;

    auto fp = from.data();
    auto tp = to.data();
    while(size--) {
        if((*fp != *tp) && (tolower(*fp) != *tp))
            return false;
        ++fp;
        ++tp;
    }
    return true;
}


auto util::get_value(std::string_view text) -> int
{
    int value = 0;
    bool negative = false;

    if(!text.empty() && text.front() == '-') {
        negative = true;
        text.remove_prefix(1);
    }

    for(auto ch : text) {
        if(ch < '0' || ch > '9')
            return 0;
        if(value > 99999999)
            return 0;
        value *= 10;
        value += (ch - '0');
    }

    if(negative)
        return -value;

    return value;
}

auto util::get_env(const char *id, size_t max) -> const char *
{
    auto cp = ::getenv(id); // FlawFinder: wrapped in verify
    if(!cp)
        return nullptr;

    if(!u8verify(cp, max))
        return nullptr;

    return cp;
}

auto util::dirname(const std::string& path) -> std::string
{
    auto end = path.rfind('/');
    return end == std::string::npos ? "." : path.substr(0, --end);
}

auto util::basename(const std::string& path, std::initializer_list<std::string_view> extensions) -> std::string
{
    auto filename = path;
    auto end = path.rfind('/');
    if(end != std::string::npos)
        filename = path.substr(++end);

    if(!extensions.size())
        return filename;

    auto size = filename.length();
    for(auto const& ext : extensions) {
        if(ext.length() > size)
            continue;
        if(filename.substr(size - ext.length()) == ext) {
            return filename.substr(0, size - ext.length());
        }
    }
    return filename;
}

auto util::exists(const std::string& path) -> bool
{
    return access(path.data(), F_OK) == 0;  // FlawFinder: ignored
}

auto util::is_dir(const std::string& path) -> bool
{
    struct stat ino{};
    return stat(path.data(), &ino) == 0 ? S_ISDIR(ino.st_mode) : false;
}

auto util::to_b64(const uint8_t *data, size_t size) -> std::string
{
    assert(size > 0);

    std::string result;
    unsigned bits{};

    while(size >= 3) {
        bits = (((unsigned)data[0])<<16) | (((unsigned)data[1])<<8)
            | ((unsigned)data[2]);
        data += 3;
        size -= 3;
        result += print_format("{}{}{}{}", alphabet[bits >> 18], alphabet[(bits >> 12) & 0x3f], alphabet[(bits >> 6) & 0x3f], alphabet[bits & 0x3f]);
    }

    if (size) {
        bits = ((unsigned)data[0])<<16;
        result += alphabet[bits >> 18];
        if (size == 1) {
            result += alphabet[(bits >> 12) & 0x3f];
            result += '=';
        }
        else {
            bits |= ((unsigned)data[1])<<8;
            result += alphabet[(bits >> 12) & 0x3f];
            result += alphabet[(bits >> 6) & 0x3f];
        }
        result += '=';
    }
    return result;
}

auto util::random_fill(uint8_t *data, size_t size) -> bool
{
    auto fd = ::open("/dev/urandom", O_RDONLY); // FlawFinder: ignore
    if(fd < 0)
        return false;

    ::read(fd, data, size); // FlawFinder: ignore
    ::close(fd);
    return true;
}

auto util::set_string(char *cp, size_t max, std::string_view view) -> size_t
{
    assert(cp != nullptr);

    auto count = view.size();
    auto dp = view.data();
    if(count >= max)
        count = max - 1;

    max = count;
    while(max--)
        *(cp++) = *(dp++);

    *cp = 0;
    return count;
}

auto util::count(const char *cp, size_t max) -> size_t
{
    size_t count = 0;
    while(cp && *cp && count < max) {
        ++count;
        ++cp;
    }
    return count;
}

auto util::get16(const uint8_t *ptr) -> uint16_t
{
    return ((uint16_t)(ptr[0]) << 8) | (uint16_t)(ptr[1]);
}

auto util::get32(const uint8_t *ptr) -> uint32_t
{
    return ((uint32_t)(ptr[0]) << 24) | ((uint32_t)(ptr[1]) << 16) |
        ((uint32_t)(ptr[2]) << 8) | (uint32_t)(ptr[3]);
}

auto util::get64(const uint8_t *ptr) -> uint64_t
{
    return ((uint64_t)(ptr[0]) << 56) | ((uint64_t)(ptr[1]) << 48) |
        ((uint64_t)(ptr[2]) << 40) | ((uint64_t)(ptr[3]) << 32) |
        ((uint64_t)(ptr[4]) << 24) | ((uint64_t)(ptr[5]) << 16) |
        ((uint64_t)(ptr[6]) << 8) | (uint64_t)(ptr[7]);
}

void util::set16(uint8_t *ptr, uint16_t value)
{
    ptr[0] = (value & 0xff00) >> 8;
    ptr[1] = value & 0xff;
}

void util::set32(uint8_t *ptr, uint32_t value)
{
    ptr[0] = (value & 0xff000000) >> 24;
    ptr[1] = (value & 0x00ff0000) >> 16;
    ptr[2] = (value & 0x0000ff00) >> 8;
    ptr[3] = value & 0xff;
}

void util::set64(uint8_t *ptr, uint64_t value)
{
    ptr[0] = (value & 0xff00000000000000) >> 56;
    ptr[1] = (value & 0x00ff000000000000) >> 48;
    ptr[2] = (value & 0x0000ff0000000000) >> 40;
    ptr[3] = (value & 0x000000ff00000000) >> 32;
    ptr[4] = (value & 0x00000000ff000000) >> 24;
    ptr[5] = (value & 0x0000000000ff0000) >> 16;
    ptr[6] = (value & 0x000000000000ff00) >> 8;
    ptr[7] = value & 0xff;
}

dir::dir() noexcept : state(nullptr)
{
}

dir::dir(dir&& from) noexcept : state(from.state)
{
    from.state = nullptr;
}

dir::dir(const std::string& path) noexcept : state(nullptr)
{
    scan(path);
}

auto dir::operator=(dir&& from) noexcept -> dir&
{
    state = from.state;
    from.state = nullptr;
    return *this;
}

dir::~dir()
{
    release();
}

void dir::release()
{
    if(!state)
        return;

    closedir(state->handle); // NOLINT
    delete state;
    state = nullptr;
}

void dir::next()
{
    if(!state)
        return;

    state->find = readdir(state->handle);
    if(!state->find)
        release();
}

void dir::scan(const std::string& path)
{
    release();
    state = new stateinfo; // NOLINT

    state->handle = opendir(path.c_str());
    state->find = nullptr;
    if(!state->handle) {
        delete state;
        state = nullptr;
        return;
    }
    next();
}

auto dir::get() -> std::string
{
    if(!state || !state->find)
        return {};
    std::string result = state->find->d_name;
    next();
    return result;
}
