/*
 * Copyright (C) 2020-2021 David Sugar <tychosoft@gmail.com>.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "compiler.hpp"
#include "ipc.hpp"

#include <print.hpp>
#include <args.hpp>
#include <audio.hpp>

#include <string>
#include <csignal>

#include <unistd.h>
#include <mqueue.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <fcntl.h>

namespace {
args::option opt_app('a',       "--app", "application script", "none");
args::option opt_lang('l',      "--lang", "language used", "en");
args::option opt_mailbox('m',   "--mailbox", "mailbox selected", "none");
args::option opt_voice('v',     "--voice", "voice used", "default");
args::option helpflag('h',      "--help", "display this list");
args::option althelp('?',       nullptr, nullptr);

auto path(const std::string& from, const audio_paths& paths)
{
    char to[256];
    if(!audio_file::path(from.c_str(), to, sizeof(to), &paths))
        throw std::runtime_error(print_format("{}: invalid path", from));
    return std::string(to);
}

auto arg(int arg, const audio_paths& paths)
{
    auto from = std::string(args::list()[arg]);
    if(from[0] == '/' || from[0] == '.') {
        if(ends_with(from, ".au"))
            return from;
        return from + ".au";
    }
    return path(from, paths);
}
} // end namespace

auto main(int argc, const char **argv) -> int   // NOLINT
{
    try {
        auto count = args::parse(argc, argv);
        if(is(helpflag) || is(althelp) || !count) {
            args::help({"bordeauxsnd [options] command..."}, {
                "bordeauxsnd " PROJECT_VERSION,
                "David Sugar <tychosoft@gmail.com>",
                "Audio file operations"
            });
            return 0;
        }

        auto command = args::list()[0];
        audio_file audio;
        audio_paths paths;

        if(is(opt_lang))
            paths.lang = *opt_lang;

        if(is(opt_mailbox))
            paths.mailbox = *opt_mailbox;

        if(is(opt_app))
            paths.app = *opt_app;

        if(is(opt_voice))
            paths.voice = *opt_voice;

        if(count == 2 && command == "path") {
            print("{}\n", path(std::string(args::list()[1]), paths));
            return 0;
        }

        if(count == 2 && command == "info") {
            auto filename = arg(1, paths);
            auto format = "unknown";
            auto bytesize = 0U;
            auto bitsize = 0U;
            auto samples = 0UL;
            if(!audio.open(filename.c_str()))   // FlawFinder: ignore
                throw std::runtime_error(print_format("{}: unknown or invalid file", filename));
            switch(audio.encoding()) {
            case 1:
                format = "g.711 u-law";
                bytesize = 1;
                break;
            case 2:
                format = "8-bit pcm";
                bytesize = 1;
                break;
            case 3:
                format = "16-bit pcm";
                bytesize = 2;
                break;
            case 4:
                format = "24-bit pcm";
                bytesize = 3;
                break;
            case 5:
                format = "32-bit pcm";
                bytesize = 4;
                break;
            case 6:
                format = "32-bit float";
                bytesize = 4;
                break;
            case 7:
                format = "64-bit float";
                bytesize = 8;
                break;
            case 8:
                format = "fragmented";
                bytesize = 0;
                break;
            case 9:
            case 21:
                format = "dsp image";
                bytesize = 0;
                break;
            case 10:
                format = "8-bit fixed";
                bytesize = 1;
                break;
            case 11:
                format = "16-bit fixed";
                bytesize = 2;
                break;
            case 12:
                format = "24-bit fixed";
                bytesize = 3;
                break;
            case 13:
                format = "32-bit fixed";
                bytesize = 4;
                break;
            case 18:
                format = "16-bit linear with emphasis";
                bytesize = 2;
                break;
            case 19:
                format = "16-bit linear compressed";
                bytesize = 2;
                break;
            case 20:
                format = "16-bit linear compressed with emphasis";
                bytesize = 2;
                break;
            case 23:
                format = "g.721 4-bit adpcm";
                bitsize = 4;
                break;
            case 24:
                format = "g.722 adpcm";
                bitsize = 4;
                break;
            case 25:
                format = "g.723 3-bit adpcm";
                bitsize = 3;
                break;
            case 26:
                format = "g.723 5-bit adpcm";
                bitsize = 5;
                break;
            case 27:
                format = "g.711 a-law";
                bytesize = 1;
                break;
            default:
                break;
            }
            if(bytesize)
                bitsize = bytesize * 8;
            bytesize *= audio.channels();
            print("path:        {}\n", filename);
            if(audio.size() == 0xffffffff)
                print("size:        unknown\n");
            else {
                print("size:        {}\n", audio.size());
                if(bitsize) {
                    samples = ((audio.size() * 8UL) / bitsize) / audio.channels();
                    samples *= 1000UL;
                    print("tine:        {} msec\n", samples / audio.sample_rate());
                }
            }
            print("encoding:    {}\n", format);
            print("channels:    {}\n", audio.channels());
            if(bitsize || bytesize)
                print("sample rate: {}\n", audio.sample_rate());
            if(bitsize)
                print("sample bits: {}\n", bitsize);
            if(!audio.annotation().empty())
                print("annotation:  {}\n", audio.annotation());
            return 0;
        }
        throw std::runtime_error(print_format("{}: unknown command or arguments"));
    }
    catch(const bad_arg& e) {
        die(-1, "Usage: bordeauxsnd [options] command...\nerror: {}\n", e.what());
    }
    catch(const std::exception& e) {
        die(-1, "error: {}\n", e.what());
    }
    catch(...) {
        die(-1, "error: unknown error\n");
    }
}

