/*
 * Copyright (C) 2020-2021 David Sugar <tychosoft@gmail.com>.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "compiler.hpp"

#include <print.hpp>
#include <strings.hpp>
#include <digest.hpp>
#include <args.hpp>

#include <termios.h>
#include <unistd.h>

using namespace crypto;

namespace {
const char digits[17] = "0123456789abcdef";

args::option opt_config('c',    "--config", "config file path");
args::option opt_digest('d',    "--digest", "algorithm to use", "sha256|md5");
args::option helpflag('h',      "--help",   "display this list");
args::option opt_passwd('p',    "--password",  "password to hash", "string");
args::option opt_realm('r',     "--realm",  "realm to encode for", "string");
args::option verbose('v',       "--verbose", "full output line");
args::option althelp('?',       nullptr, nullptr);

auto to_hex(const uint8_t *data, size_t size)
{
    std::ostringstream out;
    while(data && size--) {
        out << digits[*data >> 4] << digits[*data & 0x0f];
        ++data;
    }
    return out.str();
}

auto sha256_secret(std::string_view userid, std::string_view realm, std::string_view secret)
{
    digest_t sha(EVP_sha256());
    sha.update(userid.data(), userid.size());
    sha.update(":", 1);
    sha.update(realm.data(), realm.size());
    sha.update(":", 1);
    sha.update(secret.data(), secret.size());
    sha.finish();
    return to_hex(sha.data(), sha.size());
}

auto md5_secret(std::string_view userid, std::string_view realm, std::string_view secret)
{
    digest_t md5(EVP_md5());
    md5.update(userid.data(), userid.size());
    md5.update(":", 1);
    md5.update(realm.data(), realm.size());
    md5.update(":", 1);
    md5.update(secret.data(), secret.size());
    md5.finish();
    return to_hex(md5.data(), md5.size());
}

void set_echo(bool enable = true)
{
    struct termios tty{};
    tcgetattr(STDIN_FILENO, &tty);
    if(!enable)
        tty.c_lflag &= ~ECHO;
    else
        tty.c_lflag |= ECHO;

    (void) tcsetattr(STDIN_FILENO, TCSANOW, &tty);
}

void get_realm()
{
    static char buf[256];
    if(!gethostname(buf, sizeof(buf) - 1)) {
        buf[sizeof(buf) - 1] = 0;
        opt_realm.set(buf);
    }
    else
        throw bad_arg("no realm found or hostname set");
}

void get_pass()
{
    set_echo(false);
    std::cout << "password: ";
    std::cout.flush();
    std::string password;
    std::cin >> password;
    std::cout << std::endl;

    std::cout << "verify: ";
    std::cout.flush();
    std::string verify;
    std::cin >> verify;
    set_echo(true);
    std::cout << std::endl;

    if(verify != password)
        throw std::runtime_error("passwords do not match");

    opt_passwd.set(password.c_str());
}
} // end namespace

auto main(int argc, const char **argv) -> int
{
    int exit_code = 0;
    atexit([]{
        set_echo(true);
    });

    try {
        auto count = args::parse(argc, argv);
        if(is(helpflag) || is(althelp)) {
            args::help({"sip-hash [options] userid"}, {
                "sip-hash " PROJECT_VERSION,
                "David Sugar <tychosoft@gmail.com>",
                "Compute hash secret for a sip user"
            });
            return 0;
        }
        if(count > 1)
            throw bad_arg("only use 1 userid");

        auto userid = args::list()[0].data();

        if(!is(opt_digest))
            opt_digest.set("sha-256");

        auto digest = lower_case(*opt_digest);

        if(digest != "md5" && digest != "sha" && digest != "sha256" && digest != "sha-256") {
            digest = "unsupported algorithm " + digest;
            throw bad_arg(digest.c_str());
        }

        if(!is(opt_realm))
            get_realm();

        if(!is(opt_passwd))
            get_pass();

        if(is(verbose))
            print("{}:{}:", userid, *opt_realm);

        if(digest == "md5")
            print("{}\n", md5_secret(userid, *opt_realm, *opt_passwd));
        else
            print("{}\n", sha256_secret(userid, *opt_realm, *opt_passwd));

    }
    catch(const bad_arg& e) {
        die(-1, "Usage: sip-hash [options] userid\nerror: {}\n", e.what());
    }
    catch(const std::exception& e) {
        die(-1, "error: {}\n", e.what());
    }
    catch(...) {
        die(-1, "error: unknown error\n");
    }
    return exit_code;
}
